/*
	main.c   Cut MP3s framewise, VBR supported !
	The code is somewhat ugly, sorry! At least it is fast :-)
	(c) Jochen Puchalla 2003-2004  <mail@puchalla-online.de>
	License: GPL version 2 or later
*/

#define VERSION "1.5.1"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include <termios.h>
#include <signal.h>
#include <unistd.h>
#include <getopt.h>
#include <sys/time.h>
#include <sys/types.h>

#include "mpglib.h"

#include <readline/readline.h>
#include <readline/history.h>

#define MAXBUF 16000
/* 10 MB for VBR scanning */
#define SCANAREA 10485760
#define EOF (-1)

	unsigned int silencelength=1000, silvol=1;
	FILE *mp3file, *outfile, *timefile;
	unsigned char begin[MAXBUF], ttable[MAXBUF];
	long filesize, inpoint, outpoint;
	long startmin, endmin;
	double startsec, endsec;
	int m, s, bitrate, framesize, databegin=0;
	int fix_secondbyte=0, fix_sampfreq=0, fix_channels=7;
	double fix_frametime=0;
	int avbr, vbr=0, msec;
	char *filename, *tablename, *userin, *userout;
	int exactmode=0, mute=0, forced_file=0;
	long framenumber=0, framecount=0;
	long double totaltime=0;
	char longprefix[255]="result";
	char *oprefix=longprefix;
	char longartist[255]="unknown";
	char *artist=longartist;
	char longalbum[255]="unknown";
	char *album=longalbum;
	char longtitle[255]="";
	char *title=longtitle;
	int card=1;
	char longforcedname[255]="result.mp3";
	char *forcedname=longforcedname;

	long rememberpos;

	int debug=0;

void usage(char *s)
{
	printf("\ncutmp3  version %s  \n", VERSION);
	printf("\nUsage:  cutmp3 -i file.mp3 [-a inpoint] [-b outpoint] [-f timetable] [-o outputprefix] [-e] [-q]\n");
	printf("\n        cutmp3 -i file.mp3 -a 0:37 -b 3:57  copies the data from 0:37 to 3:57");
	printf("\n        cutmp3 -i file.mp3 -f timetable     copies the data described in timetable");
	printf("\n        cutmp3 -i file.mp3 -o song          writes the output to song01.mp3, song02.mp3,...");
	printf("\n        cutmp3 -i file.mp3 -O song.mp3      writes the output to song.mp3");
	printf("\n        cutmp3 -i file.mp3 -d 2             use second sound card");
	printf("\n\n        cutmp3 [-F] -I file.mp3             prints file information [only frames and freq.]\n\n");
	if(s) if (s[0] != '\0') printf("%s\n\n",s);
	exit(1);
}


void help(void)
{
	printf("\n\ncutmp3  version %s  \n", VERSION);
	printf("\n        Seek with 1234567890,.\n");
	printf("        Mark beginning with 'a', go there with 'A'\n");
	printf("        Mark       end with 'b', go there with 'B'\n");
	printf("        Save selection with 's'.\n");
	printf("        Save selection with ID3-tag with 't'.\n\n");
	printf("        Seek to end of next ID3-tag with 'T'\n");
	printf("        Seek to       end of next silence with 'p'\n");
	printf("        Seek to beginning of next silence with 'P'\n");
	printf("        increase volume of silence with '+'.\n");
	printf("        decrease volume of silence with '-'.\n");
	printf("        increase length of silence with 'm'.\n");
	printf("        decrease length of silence with 'n'.\n\n");
	printf("        Replay with 'r'.\n");
	printf("        Toggle sound on/off with '#'.\n");
	printf("        Write configuration file with 'S'.\n\n");
	printf("        Press 'q' to quit.\n\n");
	printf("        HINTS: 'w' is a shortcut for 'bsa'.\n");
	printf("               'W' is a shortcut for 'bta'.\n");
	return;
}


/* Which MPEG-Type ? */
/* Type is in bits nr. 11&12 */
int mpeg(unsigned char secondbyte)
{
	unsigned char mpegnumber=secondbyte;

//	printf("  mpeg()");

	mpegnumber = mpegnumber << 3;
	mpegnumber = mpegnumber >> 6;
//	printf("\nmpegnumber:%u",mpegnumber);
	if (mpegnumber==0) return 3; /* MPEG2.5 bits are 00 */
	if (mpegnumber==3) return 1; /* MPEG1   bits are 11 */
	if (mpegnumber==2) return 2; /* MPEG2   bits are 10 ; 01 is reserved */
	return 0;
}


int sampfreq(unsigned char secondbyte, unsigned char thirdbyte)
{
	unsigned char sampfreqnumber;
	int MPEG=mpeg(secondbyte);

//	printf("  sampfreq()");

	sampfreqnumber=thirdbyte;
	sampfreqnumber=sampfreqnumber<<4;
	sampfreqnumber=sampfreqnumber>>6;

	if (MPEG==1)
	{
		if (sampfreqnumber==0) return 44100;
		if (sampfreqnumber==1) return 48000;
		if (sampfreqnumber==2) return 32000;
	}
	if (MPEG==2)
	{
		if (sampfreqnumber==0) return 22050;
		if (sampfreqnumber==1) return 24000;
		if (sampfreqnumber==2) return 16000;
	}
	if (MPEG==3) /* MPEG2.5 */
	{
		if (sampfreqnumber==0) return 11025;
		if (sampfreqnumber==1) return 12000;
		if (sampfreqnumber==2) return 8000;
	}
//	printf("\nsecbyte:%u mpeg:%u sampfreqnumber:%u",secondbyte,MPEG,sampfreqnumber);
	/* In case of an error, sampfreq=1 is returned */
	return 1;
}


int channelmode(unsigned char fourthbyte)
{
//	printf("  channelmode()");

	return fourthbyte>>6;
/* result: 0=Stereo
           1=Joint-Stereo
           2=Dual-Channel
           3=Mono */
}


int isheader(int secondbyte, int thirdbyte, int fourthbyte)
{
//	printf("  isheader()");

	/* allowed:
	   111xx01x
	      00  0  226
	      00  1  227
	      10  0  242
	      10  1  243
	      11  0  250
	      11  1  251
	  not 01  0  234
	  not 01  1  235
	*/


	/* When we already know about some things: */
	if (fix_secondbyte!=0 && secondbyte!=fix_secondbyte) return 0;
	if (fix_sampfreq!=0 && sampfreq(secondbyte,thirdbyte)!=fix_sampfreq) return 0;
	if (fix_channels!=7 && channelmode(fourthbyte)!=fix_channels) return 0;

	/* second byte of mp3 header can only be 6 values  */
//	printf("\nsecondbyte=%u",secondbyte);
	if(secondbyte==226 || secondbyte==227 || secondbyte==242 || secondbyte==243 || secondbyte==250 || secondbyte==251)
	return 1;
	else
	return 0;
}

int Bitrate(unsigned char secondbyte,unsigned char thirdbyte,unsigned char fourthbyte)
{
	unsigned char bitratenumber;

//	printf("  Bitrate()");

//	printf("\nmpegnumber=%u",mpegnumber);

	/* The next section is idiot-proof ! */
	bitratenumber = thirdbyte>>4;
//	printf("\nbitratenumber=%u",bitratenumber);
	if (mpeg(secondbyte)==1)
	{	if (bitratenumber==1) return 32;
		if (bitratenumber==2) return 40;
		if (bitratenumber==3) return 48;
		if (bitratenumber==4) return 56;
		if (bitratenumber==5) return 64;
		if (bitratenumber==6) return 80;
		if (bitratenumber==7) return 96;
		if (bitratenumber==8) return 112;
		if (bitratenumber==9) return 128;
		if (bitratenumber==10) return 160;
		if (bitratenumber==11) return 192;
		if (bitratenumber==12) return 224;
		if (bitratenumber==13) return 256;
		if (bitratenumber==14) return 320;
	}
	if (mpeg(secondbyte)==2)
	{	if (bitratenumber==1) return 8;
		if (bitratenumber==2) return 16;
		if (bitratenumber==3) return 24;
		if (bitratenumber==4) return 32;
		if (bitratenumber==5) return 40;
		if (bitratenumber==6) return 48;
		if (bitratenumber==7) return 56;
		if (bitratenumber==8) return 64;
		if (bitratenumber==9) return 80;
		if (bitratenumber==10) return 96;
		if (bitratenumber==11) return 112;
		if (bitratenumber==12) return 128;
		if (bitratenumber==13) return 144;
		if (bitratenumber==14) return 160;
	}
	if (mpeg(secondbyte)==3) /* MPEG2.5 */
	{	if (bitratenumber==1) return 8;
		if (bitratenumber==2) return 16;
		if (bitratenumber==3) return 24;
		if (bitratenumber==4) return 32;
		if (bitratenumber==5) return 40;
		if (bitratenumber==6) return 48;
		if (bitratenumber==7) return 56;
		if (bitratenumber==8) return 64;
		if (bitratenumber==9) return 80;
		if (bitratenumber==10) return 96;
		if (bitratenumber==11) return 112;
		if (bitratenumber==12) return 128;
		if (bitratenumber==13) return 144;
		if (bitratenumber==14) return 160;
	}
	/* In case of an error, Bitrate=1 is returned */
	return 1;
}


int paddingbit(unsigned char thirdbyte)
{
	unsigned char temp;

//	printf("  paddingbit()");

	temp=thirdbyte<<6;
	return temp>>7;
}

int Framesize(unsigned char secondbyte,unsigned char thirdbyte,unsigned char fourthbyte)
{
	int br=Bitrate(secondbyte,thirdbyte,fourthbyte);
	int sf=sampfreq(secondbyte,thirdbyte);

//	printf("  Framesize()");
//	printf("\nFramesize() br=%u",br);
//	printf("\nFramesize() sf=%u",sf);
//	printf("\nFramesize() returnvalue=%u",144000*br/sf);
	if ((br!=1) && (sf!=1)) return (144000*br/sf)+paddingbit(thirdbyte);
	else return 1;
}


/* find next frame at seekpos */
long nextframe(long seekpos)
{
	long oldseekpos=seekpos;   /* remember seekpos */
	unsigned char a,b,c,d;

	if(seekpos>filesize) return filesize;

	fseek(mp3file, seekpos, SEEK_SET);
//	printf("seekpos=%li \n",seekpos);

	/* if seekpos is a header and Framesize is valid, jump to next header via Framesize, else move on one byte */
	a=fgetc(mp3file);
	b=fgetc(mp3file);
	c=fgetc(mp3file);
	d=fgetc(mp3file);
	if(a==255 && isheader(b,c,d)==1 && Framesize(b,c,d)!=1 && sampfreq(b,c)!=1) seekpos=seekpos+Framesize(b,c,d);
	else seekpos=oldseekpos+1;

	/* find next possible header, start right at seekpos */
	fseek(mp3file, seekpos, SEEK_SET);

	while(1)           /* loop till break */
	{
		if(seekpos>=filesize) break;
		fseek(mp3file, seekpos, SEEK_SET);
		a=fgetc(mp3file);  /* get next byte */
		if(a==255)
		{
//			printf("seekpos=%li \n",seekpos);
			b=fgetc(mp3file);
			c=fgetc(mp3file);
			d=fgetc(mp3file);
			if (isheader(b,c,d)==1 && Framesize(b,c,d)!=1 && sampfreq(b,c)!=1) break;  /* next header found */
		}
		seekpos++;
//		printf("seekpos=%li \n",seekpos);
	}

//	printf("seekpos1=%li\n",seekpos);

	/* also check next frame if possible */
	if(seekpos+Framesize(b,c,d)+4<filesize)
	{
		fseek(mp3file, seekpos+Framesize(b,c,d), SEEK_SET);
		a=fgetc(mp3file);
		b=fgetc(mp3file);
		c=fgetc(mp3file);
		d=fgetc(mp3file);
		if(a==255 && isheader(b,c,d)==1 && Framesize(b,c,d)!=1 && sampfreq(b,c)!=1);
		else seekpos=nextframe(seekpos+1);
	}

//	printf("seekpos3=%li\n",seekpos);

	return seekpos;
}


/* find previous frame at seekpos */
long prevframe(long seekpos)
{
	long oldseekpos=seekpos;   /* remember seekpos */
	long nextheaderpos;
	unsigned char a,b,c,d;

	if(seekpos<databegin)seekpos=databegin;

	/* find next header, start right at seekpos */
	nextheaderpos=nextframe(seekpos);

	/* rewind to previous header and check if nextheader is more than framesize bytes further */
	seekpos=oldseekpos-1;
	while(1)           /* loop till break */
	{
		if(seekpos<databegin)break;
		fseek(mp3file, seekpos, SEEK_SET);
		a=fgetc(mp3file);
		if(a==255)
		{
			b=fgetc(mp3file);
			c=fgetc(mp3file);
			d=fgetc(mp3file);
			if(isheader(b,c,d) && Framesize(b,c,d)!=1 && seekpos+Framesize(b,c,d)<=nextheaderpos && sampfreq(b,c)!=1)break;     /* previous frame found */
		}
		seekpos--;
//		printf("seekpos=%li \n",seekpos);
	}

	if(seekpos<databegin)seekpos=databegin;
	return seekpos;
}


int showmins (long bytes)
{
	if(bytes==-1) return totaltime/60000;
	else
	return bytes/avbr*8/60000;
}

double showsecs (long bytes)
{
	double temp;

	if(bytes==-1)
	{
		temp=totaltime/1000-showmins(-1)*60;
		if(temp>59.99) return 59.99; /* 59.997 would be displayed as 60:00 without this */
		return temp;
	}
	else
	temp=bytes;
	temp=(temp/avbr*8/1000-(showmins(bytes)*60));
	if(temp>59.99) return 59.99; /* 59.997 would be displayed as 60:00 without this */
	return temp;
}

int showintsecs (long bytes)
{
	return showsecs(bytes);
}


/*
time of frame in milliseconds is: Framesize*8/Bitrate == 1152000/sampfreq !
because Bitrate=128 means 128000bits/s
*/


long fforward (long seekpos,long skiptime)
{
	long double temptime=0;
	long double oldtotaltime=totaltime; /* remember totaltime */
	unsigned char a,b,c,d;
	long lastheaderpos=seekpos;
	int lastframesize=0;
	int foundframe=0;

//	if(exactmode==0 && skiptime==1) return nextframe(seekpos);
	if(exactmode==0) return nextframe(seekpos+msec*skiptime);

	printf("  seeking at        ");

	/* set lastframesize: */
	fseek(mp3file, seekpos, SEEK_SET);
	a=fgetc(mp3file);
	b=fgetc(mp3file);
	c=fgetc(mp3file);
	d=fgetc(mp3file);
	lastframesize=Framesize(b,c,d);

//	printf("  first temptime=%Lf   seekpos=%li     \n",temptime,seekpos);

	fseek(mp3file, seekpos, SEEK_SET);
	while(seekpos<=filesize)           /* loop till break */
	{
		if(temptime>skiptime) break;
		if(seekpos>filesize) break;
		seekpos++;
		if(foundframe){seekpos=seekpos+lastframesize-1;foundframe=0;}
		fseek(mp3file, seekpos, SEEK_SET);
		a=fgetc(mp3file);  /* get next byte */
		if(a==255)
		{
			b=fgetc(mp3file);
			c=fgetc(mp3file);
			d=fgetc(mp3file);
			if (isheader(b,c,d)==1 && Framesize(b,c,d)!=1 && seekpos>=lastheaderpos+lastframesize && sampfreq(b,c)!=1)     /* next frame found */
			{
				temptime=temptime+fix_frametime; /* increase temptime for the first time*/
				foundframe=1;
				lastheaderpos=seekpos;
				lastframesize=Framesize(b,c,d);
				if(seekpos>filesize)break;     /* EOF reached? */
				totaltime=oldtotaltime+temptime;                      /* totaltime needs to be set for exact showmins() */
				printf("\b\b\b\b\b\b\b%4u:%02u",showmins(-1),showintsecs(-1));
//				printf("  mid temptime=%Lf   seekpos=%li     \n",temptime,seekpos);
			}
//			else fseek(mp3file, seekpos, SEEK_SET);    /* rewind to next byte for if(a==255) */
		}
	}

	if(seekpos>lastheaderpos+lastframesize)temptime=temptime+fix_frametime;

	if(seekpos>filesize)
	{
		printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b  End of file reached!     \n                                        ");
		seekpos=filesize;
	}

//	printf("last  temptime=%Lf   seekpos=%li     \n",temptime,seekpos);
	totaltime=oldtotaltime+temptime;
	printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
	return seekpos;
}


long frewind (long seekpos,long skiptime)
{
	long double temptime=0;
	long double oldtotaltime=totaltime; /* remember totaltime */
	unsigned char a,b,c,d;
	long lastheaderpos;

//	if(exactmode==0 && skiptime==1) return prevframe(seekpos);
	if(exactmode==0) return prevframe(seekpos-msec*skiptime);

	lastheaderpos=seekpos+1;
	printf("  seeking at        ");

//	printf("  first temptime=%Lf   seekpos=%li     \n",temptime,seekpos);

	fseek(mp3file, seekpos, SEEK_SET);
	while(1)                              /* loop till break */
	{
		if(temptime>skiptime) break;
		seekpos--;                       /* STEP ONE BYTE BACK */
		if(seekpos<=databegin)
		{
//			printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b  Beginning of file reached!     \n                                        ");
			seekpos=databegin;
			oldtotaltime=0;
			break;
		}

		fseek(mp3file, seekpos, SEEK_SET);
		a=fgetc(mp3file);
		if(a==255)
		{
			b=fgetc(mp3file);
			c=fgetc(mp3file);
			d=fgetc(mp3file);
			if(isheader(b,c,d) && Framesize(b,c,d)!=1 && seekpos+Framesize(b,c,d)<=lastheaderpos && sampfreq(b,c)!=1)     /* previous frame found */
			{
				lastheaderpos=seekpos;
//				printf("found header\n");
				totaltime=oldtotaltime-temptime;                      /* totaltime needs to be set for exact showmins() */
				printf("\b\b\b\b\b\b\b%4u:%02u",showmins(-1),showintsecs(-1));
				if(seekpos<=databegin) break;
				temptime=temptime+fix_frametime;                      /* sum up real time */
//				printf("  mid temptime=%Lf   seekpos=%li     \n",temptime,seekpos);
		if(temptime>skiptime) break;
			}
//			else fseek(mp3file, seekpos, SEEK_SET);    /* rewind to next byte for if(a==255) */
		}
	}

//	printf("last  temptime=%Lf   seekpos=%li     \n",temptime,seekpos);
	totaltime=oldtotaltime-temptime;
	printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
	if (totaltime<0) {totaltime=0;seekpos=databegin;}
	return seekpos;
}


/* Determination of average VBR bitrate */
int avbitrate(void)
{
	long double average;
	long int pos=databegin, framecount=0, bitratesum=0;
	unsigned char a,b,c,d;
	int thisframesize=0,thisbitrate=0;
	int pad=2;

//	printf("  avbitrate");
	fseek(mp3file, databegin, SEEK_SET);
	while(1)                              /* loop till break */
	{
//		printf("\navbitrate in loop, pos=%lu, filesize=%lu\n",pos,filesize);
		if(pos>filesize-10000 || pos>SCANAREA) break;
		a=fgetc(mp3file); pos++;
//		printf("\na=%u\n",a);
		if(a==255)
		{
			b=fgetc(mp3file);
			c=fgetc(mp3file);
			d=fgetc(mp3file);
			thisframesize=Framesize(b,c,d);                                     /* cache Framesize(b,c,d) */
			if (isheader(b,c,d)==1 && thisframesize!=1 && sampfreq(b,c)!=1)     /* next frame found */
			{
				fseek(mp3file, pos-pad+thisframesize, SEEK_SET);     /* skip Framesize-pad bytes */
				pos=pos-pad+thisframesize;                           /* adjust pos to actual SEEK_SET */
				framecount++;                                        /* one more frame */
				thisbitrate=Bitrate(b,c,d);                          /* cache Bitrate(b,c,d) */
				bitratesum=bitratesum+thisbitrate;
//				printf("\nbitratesum:%lu framecount:%lu ",bitratesum,framecount);
				if(bitratesum/framecount!=thisbitrate) vbr=1;     /* if file is VBR */
//				printf("\nbitratesum:%lu framecount:%lu ",bitratesum,framecount);
			}
			else fseek(mp3file, pos, SEEK_SET);    /* rewind to next byte for if(a==255) */
		}
	}

//	printf("\navbitrate()");

	average=bitratesum/framecount;
	return average;
}


/* Determination of total number of frames */
long get_total_frames(void)
{
	long int pos=0, framecount=0;
	unsigned char a,b,c,d;
	int pad=2;
	int thisframesize=0;

	fseek(mp3file, 0, SEEK_SET);
	while(1)                              /* loop till break */
	{
		if(pos>filesize-1) break;
		a=fgetc(mp3file); pos++;
//		printf("\na=%u\n",a);
		if(a==255)
		{
			b=fgetc(mp3file);
			c=fgetc(mp3file);
			d=fgetc(mp3file);
			thisframesize=Framesize(b,c,d);                                     /* cache Framesize(b,c,d) */
			if (isheader(b,c,d)==1 && thisframesize!=1 && sampfreq(b,c)!=1)     /* next frame found */
			{
//				printf(" %li ",pos-rememberpos);
				rememberpos=pos;
				fseek(mp3file, pos-pad+thisframesize, SEEK_SET);       /* skip Framesize-pad bytes */
				pos=pos-pad+thisframesize;                             /* adjust pos to actual SEEK_SET */
				framecount++;                                          /* one more frame */
			}
			else { fseek(mp3file, pos, SEEK_SET); }    /* rewind to next byte for if(a==255) */
		}
	}
	return framecount;
}


/* This function shall return the volume (energy level)
   of one frame. Taken from mpglib and mpcut. */
unsigned int volume(long playpos)
{
	unsigned char copy[MAXBUF+5];
	FILE *volfile;
	long get_maxframe(FILE * zin);
	unsigned int level, framenum;
	int i, pos, frsize, pad;

//	printf("  volume()");

	if(debug==1)printf("\nbefore prevframe(playpos)");
	if(debug==1)printf("\nvolume playpos %li \n",playpos);
	playpos=prevframe(playpos);
	playpos=prevframe(playpos);
	playpos=prevframe(playpos);
	playpos=prevframe(playpos);
	playpos=prevframe(playpos);  /* 5-1 times prevframe because framenum=4 */
	playpos=nextframe(playpos);
	if(debug==1)printf("volume(): after 3x prevframe %li \n",playpos);

/* Copy a small section into copy[] */
	fseek(mp3file, playpos, SEEK_SET);
//	printf("\n%lu        ",playpos);
	for (i=0 ; i<MAXBUF && playpos+i<filesize ; i++)
	{
		copy[i]=fgetc(mp3file);
	}

/* Copy 6 frames to outfile */
//	system(">/tmp/volume.mp3");
	volfile = fopen("/tmp/volume.mp3","w+b"); /* open temp.vol.file read-writable*/
	pos=0;
/* There are files out there which have too big or too small frames.
   Therefore we check end of frames +-2 two bytes */
	pad=2;
	for(framecount=0; framecount < 6 ; framecount++)
	{
		/* seek next frame */
		while( ( (frsize=Framesize(copy[pos+1],copy[pos+2],copy[pos+3]))==1
		         || sampfreq(copy[pos+1],copy[pos+2])==1
		         || isheader(copy[pos+1],copy[pos+2],copy[pos+3])!=1 ) && pos < MAXBUF ) pos++; // while not header and still inside copy[], move on
//		printf("pos=%u  frsize=%u\n",pos,frsize);
		/* write frsize-pad frames */
		for(i=0; i<frsize-pad && pos<MAXBUF && playpos+pos<filesize ; i++)
		{
			fputc(copy[pos],volfile);pos++;
		}
		/* then write last bytes of frame */
		for(i=0; i<pad+2 && pos+1<MAXBUF && playpos+pos+1<filesize ; i++)
		{
			fputc(copy[pos],volfile); pos++;
//			printf("pos=%u  nextvalue=%u\n",pos,copy[pos]);
			if (copy[pos]==255 && copy[pos+1]==fix_secondbyte) // check for next frame
			{
				i=pad+2;
//				printf("pos=%u  values=%u %u %u %u %u f2nd=%u\n",pos,copy[pos],copy[pos+1],copy[pos+2],copy[pos+3],copy[pos+4],fix_secondbyte);
			}
			if (copy[pos]=='T' && copy[pos+1]=='A') // check for TAG
			{
				i=pad+2;
//				printf("pos=%u  values=%u %u %u %u %u f2nd=%u\n",pos,copy[pos],copy[pos+1],copy[pos+2],copy[pos+3],copy[pos+4],fix_secondbyte);
			}
			if (copy[pos]=='I' && copy[pos+1]=='D') // check for ID3
			{
				i=pad+2;
//				printf("pos=%u  values=%u %u %u %u %u f2nd=%u\n",pos,copy[pos],copy[pos+1],copy[pos+2],copy[pos+3],copy[pos+4],fix_secondbyte);
			}
		}
	}

/* get energy level */
	fseek(volfile, 0, SEEK_SET);
	InitMP3();
	framenum=4;
	level = get_framelevel(volfile,framenum);
	if (level==MP3_ERR) level=65535;
	if (level<0) level=(-1)*level;
	ExitMP3();
	fclose(volfile);
	return level;
}


/* This function returns the position for the next silence */
long seeksilence(long seekpos)
{
	unsigned int vol;
	long silstart, silmid;

	if(exactmode==1)
	{
		exactmode=0;
		printf("  WARNING: silence seeking switches off exact mode\n                    ");
	}
//	printf("  seeksilence()");

	seekpos=nextframe(seekpos+msec*silencelength); /* move on */

	printf("  seeking at        ");
	while (seekpos<filesize)
	{
		while ( (vol=volume(seekpos)) >= silvol && nextframe(seekpos)<filesize )
		{
			/* look for candidate for silence every silencelength milliseconds: */
			if(debug==9)printf("\ncandidate seek at %lu   volume is %u\n",seekpos,vol);
			printf("\b\b\b\b\b\b\b%4u:%02u",showmins(seekpos),showintsecs(seekpos));
			seekpos=nextframe(seekpos+msec*silencelength);
			if (nextframe(seekpos)==seekpos) return filesize; /* EOF reached? */
		}
		/* we found a candidate! */
		if(debug==9)printf("\ncandidate found at %lu   volume is %u\n",seekpos,vol);
		printf("\b\b\b\b\b\b\b%4u:%02u",showmins(seekpos),showintsecs(seekpos));
		silmid=seekpos; /* remember first found silent frame */
		seekpos=nextframe(seekpos); /* move on */

		/* rewind to first silent frame */
		while ((vol=volume(seekpos)) < silvol)
		{
			if ((seekpos=prevframe(seekpos))==seekpos) break;
			if(debug==9)printf("\nrewind seek at %lu   volume is %u\n",seekpos,vol);
			if(debug==9)printf("\ndatabegin is %u",databegin);
			if ((silmid-seekpos) >= silencelength*msec) break; /* silence already long enough */
			if (seekpos<=databegin) break;
		}

		silstart=seekpos; /* mark start of silence */
		seekpos=silmid; /* go back to first found silent frame */
		if(debug==9)printf("\nsilmid at %lu   volume is %u\n",seekpos,vol);

		/* quick check ! */
		if (volume(silstart+msec*silencelength) >= silvol)
			seekpos=seekpos+msec*silencelength;
		else /* scan length of silence */
		{
			/* Now we have the first quiet frame, do a length inquiry for silence
			This is not very fast... */
			while ((vol=volume(seekpos)) < silvol)
			{
				if(debug==9)printf("\nsilstart is %lu   seek at nextframe: %lu   volume is %u                ",silstart,nextframe(seekpos),vol);
				printf("\b\b\b\b\b\b\b%4u:%02u",showmins(seekpos),showintsecs(seekpos));
				if ((seekpos=nextframe(seekpos)) == nextframe(seekpos)) return filesize; /* move on and EOF reached? */
				if (seekpos+msec*silencelength>=filesize) return filesize; /* distance to EOF shorter than minimum silencelength? */
				if ((seekpos-silstart) >= silencelength*msec*15) /* silence much longer than wanted */
				{
					printf("\n\nSilence is already much longer than wanted!\nPress p to search on.                    ");
					return nextframe(seekpos-silencelength*msec*2);
				}
			}
			if ((seekpos-silstart) >= silencelength*msec) /* silence long enough? */
			{
				return prevframe(seekpos-(msec/2*silencelength)); /* short silence in the beginning */
			}
		}
	}
	return filesize; /* EOF */
}


/* This function returns the position of the beginning for the next silence */
long seeksilstart(long seekpos)
{
	unsigned int vol;
	long silstart, silmid;

	if(exactmode==1)
	{
		exactmode=0;
		printf("  WARNING: silence seeking switches off exact mode\n                    ");
	}
//	printf("  seeksilence()");

	seekpos=nextframe(seekpos+msec*silencelength); /* move on */

	printf("  seeking at        ");
	while (seekpos<filesize)
	{
		while ( (vol=volume(seekpos)) >= silvol && nextframe(seekpos)<filesize )
		{
			/* look for candidate for silence every silencelength milliseconds: */
			if(debug==9)printf("\ncandidate seek at %lu   volume is %u\n",seekpos,vol);
			printf("\b\b\b\b\b\b\b%4u:%02u",showmins(seekpos),showintsecs(seekpos));
			seekpos=nextframe(seekpos+msec*silencelength);
			if (nextframe(seekpos)==seekpos) return filesize; /* EOF reached? */
		}
		/* we found a candidate! */
		if(debug==9)printf("\ncandidate found at %lu   volume is %u\n",seekpos,vol);
		printf("\b\b\b\b\b\b\b%4u:%02u",showmins(seekpos),showintsecs(seekpos));
		silmid=seekpos; /* remember first found silent frame */
		seekpos=nextframe(seekpos); /* move on */

		/* rewind to first silent frame */
		while ((vol=volume(seekpos)) < silvol)
		{
			if ((seekpos=prevframe(seekpos))==seekpos) break;
			if(debug==9)printf("\nrewind seek at %lu   volume is %u\n",seekpos,vol);
			if(debug==9)printf("\ndatabegin is %u",databegin);
			if (seekpos<=databegin) break;
		}

		silstart=seekpos; /* mark start of silence */
		if(debug==9)printf("\nsilmid at %lu   volume is %u\n",seekpos,vol);

		/* quick check ! */
		if (volume(silstart+msec*silencelength) >= silvol)
			seekpos=seekpos+msec*silencelength;
		else /* scan length of silence */
		{
			/* Now we have the first quiet frame, do a length inquiry for silence
			This is not very fast... */
			while ((vol=volume(seekpos)) < silvol && seekpos-silstart < silencelength*msec) /* don't seek further than silencelength here */
			{
				if(debug==9)printf("\nsilstart is %lu   seek at nextframe: %lu   volume is %u                ",silstart,nextframe(seekpos),vol);
				printf("\b\b\b\b\b\b\b%4u:%02u",showmins(seekpos),showintsecs(seekpos));
				if ((seekpos=nextframe(seekpos)) == nextframe(seekpos)) return filesize; /* move on and EOF reached? */
				if (seekpos+msec*silencelength>=filesize) return filesize; /* distance to EOF shorter than minimum silencelength? */
			}
			if ((seekpos-silstart) >= silencelength*msec) /* silence long enough? */
			{
				return prevframe(silstart+(msec/2*silencelength)); /* return end of noise plus short silence */
			}
		}
	}
	return filesize; /* EOF */
}


/* This function prints the reached ID3 v1 TAG */
void showtag(long seekpos)
{
	unsigned char i;

	fseek(mp3file, seekpos, SEEK_SET); /* set to right position */
	printf("\n\nReached end of this track:   ");
	for (i=0; i < 3; i++) fgetc(mp3file); /* read 'TAG' */
	printf("\nTitle:   ");
	fgets(longtitle,31,mp3file); /* save title for later */
	printf("%s",title);
	printf("\nArtist:  ");
	fgets(longartist,31,mp3file); /* save artist for later */
	printf("%s",artist);
	printf("\nAlbum:   ");
	for (i=0; i < 30; i++) printf("%c",fgetc(mp3file));
	printf("\nYear:    ");
	for (i=0; i < 4; i++) printf("%c",fgetc(mp3file));
	printf("\nComment: ");
	for (i=0; i < 28; i++) printf("%c",fgetc(mp3file));
	printf("\n\n[1..0,r,a,b,s,q] >                      ");
	return;
}


/* This function returns the position for the next ID3 v1 TAG */
long seektag(long seekpos)
{
	unsigned char a,b,c,tag=0;

	if(exactmode==1)
	{
		exactmode=0;
		printf("  WARNING: tag seeking switches off exact mode\n                    ");
	}

//	printf("%lu\n",seekpos);
	printf("  seeking at        ");
	while (seekpos<filesize)
	{
		fseek(mp3file, seekpos, SEEK_SET); /* (re)set to right position */
		while ((a=fgetc(mp3file))!='T' && seekpos<filesize) {seekpos++;} /* look for 'T' */
		seekpos++;
		printf("\b\b\b\b\b\b\b%4u:%02u",showmins(seekpos),showintsecs(seekpos));
		b=fgetc(mp3file);
		c=fgetc(mp3file);
		if (a=='T' && b=='A' && c=='G' ) /* TAG is there */
		{
			if (nextframe(seekpos)==seekpos+127) tag=1; /* check if TAG v1 is real, next frame must be OK or TAG v2 */
			fseek(mp3file, seekpos+127, SEEK_SET);
			a=fgetc(mp3file);
			b=fgetc(mp3file);
			c=fgetc(mp3file);
			if (a=='I' && b=='D' && c=='3' ) tag=1; /* TAG v2 is next */
			if (tag==1) {showtag(seekpos-1); return seekpos+127;} /* TAG v1 is real */
		}
	}
	return filesize; /* no more TAGS */
}


/* This function writes the configuration file */
void writeconf(void)
{
	FILE *conffile;
	char filename1[255];

	sprintf(filename1,"%s/%s",getenv("HOME"),".cutmp3rc");
	/* check for conffile */
	if(NULL== (conffile = fopen(filename1,"wb"))) {printf("  configuration file could NOT be written!     \n");return;}

	fprintf(conffile, "leng=");
	fprintf(conffile, "%u",silencelength);
	fprintf(conffile, "\n");
	fprintf(conffile, "volu=");
	fprintf(conffile, "%u",silvol);
	fprintf(conffile, "\n");
	fprintf(conffile, "card=");
	fprintf(conffile, "%u",card);
	fprintf(conffile, "\n");
	fprintf(conffile, "mute=");
	fprintf(conffile, "%u",mute);
	fprintf(conffile, "\n");
	fclose(conffile);
	printf("  configuration file has been written.     \n");
	return;
}


/* This function plays sound at the current position */
void playsel(long int playpos)
{
	long bytesin;
	char outname[13]="/tmp/cut.mp3";
	long pos;

	if (!mute)
	{
		system(">/tmp/cut.mp3");
		if(playpos < 0)playpos=databegin;
		if(playpos > filesize)playpos=filesize;
		outfile = fopen(outname,"wb");
		fseek(mp3file, playpos, SEEK_SET);
		for(bytesin=0; bytesin < avbr*128; bytesin++) {fputc(getc(mp3file),outfile);}
		fclose(outfile);
	}

	if (exactmode==0) pos=playpos; else pos=-1;

	if (playpos >= inpoint)
	printf("  playing at %u:%05.2f / ~%u:%02u  (~%u:%02u after startpoint) \n",showmins(pos),showsecs(pos),m,s,showmins(playpos-inpoint),showintsecs(playpos-inpoint));
	else
	printf("  playing at %u:%05.2f / ~%u:%02u  (~%u:%02u before startpoint) \n",showmins(pos),showsecs(pos),m,s,showmins(inpoint-playpos),showintsecs(inpoint-playpos));

	if (!mute)
	{
		if (card==2)
		system("mpg123 /tmp/cut.mp3 -a /dev/dsp1 >& /tmp/mpg123.log &");
		else if (card==3)
		system("mpg123 /tmp/cut.mp3 -a /dev/dsp2 >& /tmp/mpg123.log &");
		else if (card==4)
		system("mpg123 /tmp/cut.mp3 -a /dev/dsp3 >& /tmp/mpg123.log &");
		else if (card==5)
		system("mpg123 /tmp/cut.mp3 -a /dev/dsp4 >& /tmp/mpg123.log &");
		else
		system("mpg123 /tmp/cut.mp3 >& /tmp/mpg123.log &");
	}
}


/* This function reads the configuration file */
void readconf(void)
{
	FILE *conffile;
	long confsize=0,pos=0;
	unsigned char conf[MAXBUF];
	char filename1[255];

	sprintf(filename1,"%s/%s",getenv("HOME"),".cutmp3rc");      /* get name of conffile */
	if(NULL == (conffile = fopen(filename1,"rb"))) return;      /* check for conffile */
	fseek(conffile, 0, SEEK_END);
	confsize=ftell(conffile);         /* size of conffile */

	/* read conffile into buffer: */
	fseek(conffile, 0, SEEK_SET);
	for (pos=0 ; pos<MAXBUF && pos<confsize ; pos++) conf[pos]=fgetc(conffile);
	fclose(conffile);

	pos=0;        /* reset pos */
	while(1)                              /* loop till break */
	{
		if(pos==confsize || pos==MAXBUF) break;
		if(conf[pos]=='l')
		{
			pos++;
			if (conf[pos]=='e' && conf[pos+1]=='n' && conf[pos+2]=='g' && conf[pos+3]=='=')
			{
				pos=pos+4;
				silencelength=0;
				while(isdigit(conf[pos]))
				{
					silencelength = silencelength*10 + conf[pos]-48;
					pos++;
				}
			}
		}
		if(conf[pos]=='v')
		{
			pos++;
			if (conf[pos]=='o' && conf[pos+1]=='l' && conf[pos+2]=='u' && conf[pos+3]=='=')
			{
				pos=pos+4;
				silvol=0;
				while(isdigit(conf[pos]))
				{
					silvol = silvol*10 + conf[pos]-48;
					pos++;
				}
			}
		}
		if(conf[pos]=='c')
		{
			pos++;
			if (conf[pos]=='a' && conf[pos+1]=='r' && conf[pos+2]=='d' && conf[pos+3]=='=')
			{
				pos=pos+4;
				if(isdigit(conf[pos]))
				{
					card = conf[pos]-48;
					pos++;
				}
			}
		}
		if(conf[pos]=='m')
		{
			pos++;
			if (conf[pos]=='u' && conf[pos+1]=='t' && conf[pos+2]=='e' && conf[pos+3]=='=')
			{
				pos=pos+4;
				if(isdigit(conf[pos]))
				{
					mute = conf[pos]-48;
					pos++;
				}
			}
		}
		pos++;
	}
	return;
}

void playfirst(long int playpos)
{
	printf("\n1 <- 5 seeks backward");
	printf("\n6 -> 0 seeks forward");
	printf("\n ,  .  rewinds/skips one frame");
	printf("\n     a sets startpoint, A goes there");
	printf("\n     b sets endpoint, B goes there");
	printf("\n     s saves");
	printf("\n     p seeks to next silence");
	printf("\n\n     length of silence is %u milliseconds",silencelength);
	printf("\n     maximum volume during silence is %u ",silvol);
	printf("\n     using soundcard #%u \n",card);
	printf("\n     Press 'h' for more help\n");
	printf("\n[1..0,r,a,b,s,q] >  ");
	playsel(playpos);
}


void savesel(void)
{
	int zehner;
	char outname[1024]="\0";
	char *sep=" - ";
	char *suffix=".mp3";
	char *number="01";
	int count=1, i;
	long bytesin=0;
//	long oldbytes=0;
	size_t prefixlength=strlen(oprefix);
	FILE *fp;

	if(strlen(longtitle)>0) /* title known from tag? */
	{
		for(i=29 ; longartist[i]==' ' ; i--)longartist[i]='\0';
		strcat(outname, artist);
		strcat(outname, sep);
		for(i=29 ; longtitle[i]==' ' ; i--)longtitle[i]='\0';
		strcat(outname, title);
		strcat(outname, suffix);
	}
	else /* title not known from tag */
	{
		/* outname = oprefix number suffix */
		strcat(outname, oprefix);
		strcat(outname, number);
		strcat(outname, suffix);

		if(inpoint < 0) inpoint=databegin;
		if(outpoint > filesize) outpoint=filesize;
		if(outpoint < inpoint)
		{printf("  endpoint (%u:%05.2f) must be after startpoint (%u:%05.2f)!  \n",showmins(inpoint),showsecs(inpoint),showmins(outpoint),showsecs(outpoint));return;}
		if(outpoint == inpoint)
		{printf("  endpoint must be after startpoint!  \n");return;}

		/* if file exists, increment number */
		fp = fopen(outname,"r");
		while(fp != NULL)
		{
			count++;
			fclose(fp);
			if(count>99) usage("99 files written. Please choose another prefix via '-o prefix'");
			zehner=count/10;
			outname[prefixlength]=zehner+48; /* 10er-Stelle */
			outname[prefixlength+1]=(count-zehner*10)+48; /* 1er-Stelle */
			fp = fopen(outname,"r");
		}
	}

	longtitle[0]='\0'; /* erase title */

	if(forced_file==1)
	{
	outname[0]='\0';
	strcat(outname, forcedname);
	}

	outfile = fopen(outname,"wb");

	fseek(mp3file, inpoint, SEEK_SET);
	for(bytesin=0 ; bytesin < outpoint-inpoint ; bytesin++)
	{
		fputc(getc(mp3file),outfile);
//		if(bytesin-oldbytes==1024){printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b  wrote %li KiloBytes",bytesin/1024); oldbytes=bytesin;};
	}
//	printf("\n");
	fclose(outfile);
	/* noninteractive cutting: */
	if(tablename!=0) printf("  saved %lu:%05.2f - %lu:%05.2f to '%s'.  \n",startmin,startsec,endmin,endsec,outname);
	else
	/* interactive cutting: */
	printf("  saved %u:%05.2f - %u:%05.2f to '%s'.  \n",showmins(inpoint),showsecs(inpoint),showmins(outpoint),showsecs(outpoint),outname);
	return;
}



void savewithtag(void)
{
	int i;
	char comment[32]="created with cutmp3";
	char outname[1024]="cutmp3.tmp";
	char *oldname=outname;
	char outname2[1024]="\0";
	char *newname=outname2;
	char sep[4]=" - ";
	char suffix[5]=".mp3";
	long bytesin;
	signed int genre=-1;
	int count=1;
	char longtmp[255]="";
	char *tmp=longtmp;

	if(inpoint < 0) inpoint=databegin;
	if(outpoint > filesize) outpoint=filesize;
	if(outpoint <= inpoint)
		{printf("  endpoint (%u:%05.2f) must be after startpoint (%u:%05.2f)!  \n",showmins(inpoint),showsecs(inpoint),showmins(outpoint),showsecs(outpoint));return;}


//	printf("before file open in savewithtag\n");
	outfile = fopen(outname,"wb");
	fseek(mp3file, inpoint, SEEK_SET);
	for(bytesin=0; bytesin < outpoint-inpoint; bytesin++) {fputc(getc(mp3file),outfile);}

	/* write tag here */
	fputs("TAG",outfile);

	tmp=readline("\n\nTitle? ");
	tmp[30]='\0';
	title=tmp;
	fprintf(outfile,tmp);
	for(i=strlen(tmp) ; i<30 ; i++) fputc(32,outfile);

	printf("\nPress <ENTER> for '%s'",artist);
	tmp=readline("\nArtist? ");
	if(strlen(tmp)>0) artist=tmp; /* recycle old artist name when hitting ENTER */
	else tmp=artist;
	tmp[30]='\0';
	strcat(newname, artist);
	strcat(newname, sep);
	strcat(newname, title);
	strcat(newname, suffix);
	fprintf(outfile,tmp);
	for (i=strlen(tmp) ; i<30 ; i++) fputc(32,outfile);

	printf("\nPress <ENTER> for '%s'",album);
	tmp=readline("\nAlbum? ");
	if(strlen(tmp)>0) album=tmp; /* recycle old album name when hitting ENTER */
	else tmp=album;
	tmp[30]='\0';
	fprintf(outfile,tmp);
	for (i=strlen(tmp) ; i<30 ; i++) fputc(32,outfile);

	tmp=readline("\nYear? ");
	tmp[4]='\0';
	fprintf(outfile,tmp);
	for (i=strlen(tmp) ; i<4 ; i++) fputc(32,outfile);

	tmp=readline("\nComment? ");
	if (strlen(tmp)==0) tmp=comment;
	tmp[28]='\0';
	fprintf(outfile,tmp);
	for (i=strlen(tmp) ; i<28 ; i++) fputc(32,outfile);

	fputc(0,outfile);
	fputc(count,outfile);
	fputc(genre,outfile);

	fclose(outfile);

	if(forced_file==1)
	newname=forcedname;

	rename(oldname,newname);

//	printf("  inpoint %lu   outpoint %lu   filesize %lu\n",inpoint,outpoint,filesize);
	printf("\n  saved %u:%02.0f - %u:%02.0f with ID3-tag to '%s'.  \n",showmins(inpoint),showsecs(inpoint),showmins(outpoint),showsecs(outpoint),newname);
	return;
}




/* ################### KEY PRESS ######################## */

static struct termios  current,  /* new terminal settings             */
                       initial;  /* initial state for restoring later */

/* Restore term-settings to those saved when term_init was called */
void  term_restore  (void)  {
  tcsetattr(0, TCSANOW, &initial);
}  /* term_restore */

/* Clean up terminal; called on exit */
void  term_exit  ()  {
	term_restore();
	printf("\n\nbugreports to mail@puchalla-online.de\n\n");
	exit(0);
}	/* term_exit */

/* Will be called when ctrl-z is pressed, this correctly handles the terminal */
void  term_ctrlz  ()  {
	signal(SIGTSTP, term_ctrlz);
	term_restore();
	kill(getpid(), SIGSTOP);
}  /* term_ctrlz */

/* Will be called when application is continued after having been stopped */
void  term_cont  ()  {
	signal(SIGCONT, term_cont);
	tcsetattr(0, TCSANOW, &current);
}  /* term_cont */

/* Needs to be called to initialize the terminal */
void  term_init  (void)  {
	/* if stdin isn't a terminal this fails. But    */
	/* then so does tcsetattr so it doesn't matter. */

	tcgetattr(0, &initial);
	current = initial;                      /* Save a copy to work with later  */
	signal(SIGINT,  term_exit);             /* We _must_ clean up when we exit */
	signal(SIGQUIT, term_exit);
	signal(SIGTSTP, term_ctrlz);            /* Ctrl-Z must also be handled     */
	signal(SIGCONT, term_cont);
	//atexit(term_exit);
}	/* term_init */

/* Set character-by-character input mode */

void  term_character  (void)  {
	/* One or more characters are sufficient to cause a read to return */
	current.c_cc[VMIN]   = 1;
	current.c_cc[VTIME]  = 0;  /* No timeout; read forever until ready */
	current.c_lflag     &= ~ICANON;           /* Line-by-line mode off */
	tcsetattr(0, TCSANOW, &current);
}  /* term_character */

/* Return to line-by-line input mode */

void  term_line  (void)  {
	current.c_cc[VEOF]  = 4;
	current.c_lflag    |= ICANON;
	tcsetattr(0, TCSANOW, &current);
}  /* term_line */

/* Key pressed ? */

int  kbhit  (void)  {
	struct timeval  tv;
	fd_set          read_fd;

	/* Do not wait at all, not even a microsecond */
	tv.tv_sec  = 0;
	tv.tv_usec = 0;

	/* Must be done first to initialize read_fd */
	FD_ZERO(&read_fd);

	/* Makes select() ask if input is ready;  0 is file descriptor for stdin */
	FD_SET(0, &read_fd);

	/* The first parameter is the number of the largest fd to check + 1 */
	if (select(1, &read_fd,
		NULL,         /* No writes        */
		NULL,         /* No exceptions    */
		&tv)  == -1)
	return(0);                 /* an error occured */

	/* read_fd now holds a bit map of files that are readable. */
	/* We test the entry for the standard input (file 0).      */
	return(FD_ISSET(0, &read_fd)  ?  1  :  0);
}

/* ------------------------------------------------------------------------- */

void cutexact(void)
{
	long pos=databegin;
	unsigned char a,b,c,d;
	long double time=0;
	double sf=0;

	if(startmin*60+startsec >= endmin*60+endsec) {outpoint=inpoint=0; savesel(); return;}

	fseek(mp3file, databegin, SEEK_SET);
	while(1)                              /* loop till break */
	{
		if(time>=startmin*60+startsec) break;       /* inpoint reached */
		if(pos>=filesize) break;                    /* EOF reached */
		a=fgetc(mp3file); pos++;
		if(a==255)
		{
			b=fgetc(mp3file);
			c=fgetc(mp3file);
			d=fgetc(mp3file);
			if (isheader(b,c,d)==1 && Framesize(b,c,d)!=1)     /* next frame found */
			{
//				printf("found header\n");
				fseek(mp3file, pos-1+Framesize(b,c,d), SEEK_SET);    /* skip Framesize bytes */
				pos=pos-1+Framesize(b,c,d);                          /* adjust pos to actual SEEK_SET */
				if(sf==0) sf=sampfreq(b,c);                          /* do we know sampfreq? */
				time=time+(1152/sf);                                 /* sum up real time, time of one frame in seconds is 1152/sf */
			}
			else fseek(mp3file, pos, SEEK_SET);    /* rewind to next byte for if(a==255) */
		}
	}
	inpoint=pos;
/* inpoint reached, now search for outpoint */

	while(1)                              /* loop till break */
	{
		/* pos must point to end of header in case next if() is true */
		if(time>=endmin*60+endsec) break;       /* outpoint reached */
		if(pos>=filesize) break;                /* EOF reached */
		a=fgetc(mp3file); pos++;
		if(a==255)
		{
			b=fgetc(mp3file);
			c=fgetc(mp3file);
			d=fgetc(mp3file);
			if (isheader(b,c,d)==1 && Framesize(b,c,d)!=1)     /* next frame found */
			{
				fseek(mp3file, pos-1+Framesize(b,c,d), SEEK_SET);    /* skip Framesize bytes */
				pos=pos-1+Framesize(b,c,d);                          /* adjust pos to actual SEEK_SET */
				if(sf==0) sf=sampfreq(b,c);                          /* do we know sampfreq? */
				time=time+(1152/sf);                                 /* sum up real time */
			}
			else fseek(mp3file, pos, SEEK_SET);    /* rewind to next byte for if(a==255) */
		}
	}
	outpoint=pos;
/* outpoint reached, now write file */

	savesel();
//	printf("%u %u %u %u",startsec,endsec,inpoint,outpoint);
	startsec=startmin=endsec=endmin=0;
	return;
}


void writetable(void)
{	outfile = fopen("/tmp/timetable","wb");
	fprintf(outfile, userin);
	fprintf(outfile, " ");
	fprintf(outfile, userout);
	fprintf(outfile, "\n");
	fclose(outfile);
	return;
}

void cutfromtable(void)
{	int pos2, position=1;
	double number;
	startsec=startmin=endsec=endmin=0;

	timefile = fopen(tablename,"rb");
	/* Load MAXBUF Bytes into unsigned buffer[] */
	for (pos2=0 ; pos2<MAXBUF ; pos2++) ttable[pos2]=fgetc(timefile);
	fclose(timefile);

	pos2=0;
	do
	{
		/* position: startmin=1 startsek=2 startten=3 starthun=4 */
		/*             endmin=5   endsek=6   endten=7   endhun=8 */
		number=ttable[pos2]-48;
		if (isdigit(ttable[pos2]) && position==1) { startmin=startmin*10+ttable[pos2]-48; }
		if (isdigit(ttable[pos2]) && position==2) { startsec=startsec*10+ttable[pos2]-48; }
		if (isdigit(ttable[pos2]) && position==3) { startsec=startsec+number/10; position++; pos2++; number=ttable[pos2]-48; }
		if (isdigit(ttable[pos2]) && position==4) { startsec=startsec+number/100; }
		if (isdigit(ttable[pos2]) && position==5) { endmin=endmin*10+ttable[pos2]-48; }
		if (isdigit(ttable[pos2]) && position==6) { endsec=endsec*10+ttable[pos2]-48; }
		if (isdigit(ttable[pos2]) && position==7) { endsec=endsec+number/10; position++; pos2++; number=ttable[pos2]-48; }
		if (isdigit(ttable[pos2]) && position==8) { endsec=endsec+number/100; }

		/* ':' and '.' increase position */
		if (ttable[pos2]==58) position++;
		if (ttable[pos2]==46) position++;

		/* space switches to read end or to cutexact */
		if (isspace(ttable[pos2]) && position<5 && position>1) position=5;
		if (isspace(ttable[pos2]) && position>5)
		{
			cutexact();
			position=1;
			startsec=startmin=endsec=endmin=0;
		}

		if (ttable[pos2]==255) pos2=MAXBUF;
//		printf("pos2=%u %lu:%02.2f %lu:%02.2f position=%u\n",pos2,startmin,startsec,endmin,endsec,position);
		pos2++;
	}
	while (pos2<MAXBUF);
}

/* ------------------------------------------------------------------------- */
/*      */
/* MAIN */
/*      */
/* ------------------------------------------------------------------------- */

int main(int argc, char *argv[])
{
	int c,char1,startpos,showinfo=0,framestat=0;
	long int pos;

/* ------------------------------------------------------------------------- */
/*                   */
/* parse commandline */
/*                   */
/* ------------------------------------------------------------------------- */

	readconf();       /* read configuration file first! */

	if (argc<2) usage("Error: missing filename");
	else
	{
		while ((char1 = getopt(argc, argv, "f:a:b:i:o:O:I:d:heqF")) != -1)
		{
			switch (char1)
			{
			case 'a':
				if(optarg!=0) userin=optarg;
//				else usage("Error: missing time argument");
				break;
			case 'b':
				if(optarg!=0) userout=optarg;
//				else usage("Error: missing time argument");
				break;
			case 'd':
				if(optarg!=0) card=atoi(optarg);
				break;
			case 'f':
				if(NULL == (timefile = fopen(optarg,"rb") ) ) { perror(optarg); exit(98); }
				else tablename=optarg;
				break;
			case 'h':
				usage("");
				break;
			case 'e':
				exactmode=1;
				break;
			case 'q':
				mute=1;
				break;
			case 'i':
//				if(optarg==0) usage("Error: option requires an argument");
				if(NULL == (mp3file = fopen(optarg,"rb") ) ) { perror(optarg); exit(99); }
				else filename=optarg;
				break;
			case 'I':
//				if(optarg==0) usage("Error: option requires an argument");
				if(NULL == (mp3file = fopen(optarg,"rb") ) ) { perror(optarg); exit(99); }
				else { filename=optarg; showinfo=1; }
				break;
			case 'F':
				framestat=1;
				break;
			case 'o':
				if(optarg!=0) oprefix=optarg;
//				else usage("Error: missing outputprefix");
				break;
/* ThOr: let user define exactly one output file */
			case 'O':
				if(optarg!=0) { forcedname=optarg; forced_file=1; }
				break;
			default:
				exit(1);
				break;
			}
		}
	}

	if(userin==0 && userout!=0) usage("Error: missing inpoint");
	if(userin!=0 && userout==0) usage("Error: missing outpoint");
	if (filename==0) usage("Error: missing filename, use 'cutmp3 -i file.mp3 [options]'");

	setvbuf(stdout,NULL,_IONBF,0); /* switch off buffered output to stdout */

/************************************/
/********** OPEN FILE ***************/
/************************************/
	mp3file = fopen(filename, "rb");


/* get file size */
	fseek(mp3file, 0, SEEK_END);
	filesize=ftell(mp3file);
//  printf("Size : %ld\n", filesize);

/* filesize should be at least MAXBUF Bytes */
	if (filesize<MAXBUF) {printf("\n%s is too short\n\n",filename); exit(99);}

	fseek(mp3file, 0, SEEK_SET);
	/* Load MAXBUF Bytes into unsigned begin[] */
	for (pos=0 ; pos<MAXBUF ; pos++) begin[pos]=fgetc(mp3file);

/* get starting position */
	startpos=inpoint=databegin=nextframe(0);
	outpoint=filesize;

	/* global values: */
	fix_secondbyte=begin[databegin+1];
	fix_sampfreq=sampfreq(begin[databegin+1],begin[databegin+2]);
	fix_channels=channelmode(begin[databegin+3]);
	fix_frametime=1152000/fix_sampfreq;

	framesize=Framesize(begin[databegin+1],begin[databegin+2],begin[databegin+3]);
	bitrate=Bitrate(begin[databegin+1],begin[databegin+2],begin[databegin+3]);
//	printf("\nBitrate:%u Framesize:%u",bitrate,framesize);

	avbr=avbitrate(); //average bitrate
//	printf("\nVBR:%u",avbr);

	msec=avbr/8;  /* one millisecond */
//	printf("\nmsec=%u",msec);

	/* Show file info only? */
	if(showinfo==1)
	{
		if(framestat==1)
		{
		printf("%li %u\n",get_total_frames(),sampfreq(begin[databegin+1],begin[databegin+2]));
		exit(0);
		}

		printf("\n* Properties of \"%s\":\n* \n",filename);
		if (databegin != 0)
		printf("* Data starts at %u Bytes\n",databegin);
		printf("* Size of first frame is %u Bytes\n",framesize);
		/*  printf("* Framesize is %u Bytes\n",framesize-(framesize%2)); */
		if (mpeg(begin[databegin+1])==3)printf("* File is MPEG2.5-Layer3\n");
		else printf("*\n* Format is MPEG%u-Layer3\n",mpeg(begin[databegin+1]));
		if (vbr==0) printf("*           %u kBit/sec\n",avbr);
		else printf("*           %u kBit/sec (VBR average)\n",avbr);
		printf("*           %u Hz ",sampfreq(begin[databegin+1],begin[databegin+2]));
		if (channelmode(begin[databegin+3])==0) printf("Stereo\n");
		if (channelmode(begin[databegin+3])==1) printf("Joint-Stereo\n");
		if (channelmode(begin[databegin+3])==2) printf("Dual-Channel\n");
		if (channelmode(begin[databegin+3])==3) printf("Mono\n");
		if (vbr==0) printf("*\n* Filelength is %u minute(s) %4.2f second(s)\n\n",showmins(filesize),showsecs(filesize));
		else printf("*\n* Filelength is approximately %u minute(s) \n\n",showmins(filesize));
		exit(0);
	}

	/* -a and -b used? */
	if(userin!=0 && userout!=0) {tablename="/tmp/timetable"; writetable(); cutfromtable(); exit(0);}

	/* timetable used? */
	if(tablename!=0) {cutfromtable(); exit(0);}

	/* get total mins:secs */
	m=showmins(filesize);
	s=showintsecs(filesize);

	term_init();
	term_character();

	playfirst(databegin);

	while (1)
	{	printf("\n[1..0,r,a,b,s,q] > ");
//  if((c = getchar()) >= ' ');
		c = getchar();
		if(c == 'q') {printf("\n\n");exit(0);}
		if(c == '1') {startpos=frewind(startpos,1000*600); playsel(startpos);}
		if(c == '2') {startpos=frewind(startpos,1000*60); playsel(startpos);}
		if(c == '3') {startpos=frewind(startpos,1000*10); playsel(startpos);}
		if(c == '4') {startpos=frewind(startpos,1000); playsel(startpos);}
		if(c == '5') {startpos=frewind(startpos,100); playsel(startpos);}
		//previous frame, when at EOF
		if(c == ',') {startpos=frewind(startpos,1); playsel(startpos);}
//		if(c == '.') {if(nextframe(startpos)!=startpos)totaltime=totaltime+fix_frametime; startpos=nextframe(startpos); playsel(startpos);}    //next frame
		if(c == '.') {startpos=fforward(startpos,1); playsel(startpos);}    //next frame
		if(c == '6') {startpos=fforward(startpos,100);playsel(startpos);}
		if(c == '7') {startpos=fforward(startpos,1000);playsel(startpos);}
		if(c == '8') {startpos=fforward(startpos,1000*10);playsel(startpos);}
		if(c == '9') {startpos=fforward(startpos,1000*60);playsel(startpos);}
		if(c == '0') {startpos=fforward(startpos,1000*600);playsel(startpos);}
		if(c == 'a') {inpoint=startpos;printf("  startpoint set to %u:%05.2f  \n",showmins(inpoint),showsecs(inpoint));}
		if(c == 'b') {outpoint=startpos;printf("  endpoint set to %u:%05.2f  \n",showmins(outpoint),showsecs(outpoint));}
		if(c == 'r') {playsel(startpos);}
		if(c == 'v') {printf("  volume=%u",volume(startpos));}
		if(c == '#') {mute=1-mute; if(mute)printf("  sound off\n");else printf("  sound on\n");}
		if(c == 'p') {startpos=seeksilence(startpos);
		              printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
		              if (startpos>=prevframe(filesize))
		              printf("  End of file reached     \n");
		              else playsel(startpos);}
		if(c == 'P') {startpos=seeksilstart(startpos);
		              printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
		              if (startpos>=prevframe(filesize))
		              printf("  End of file reached     \n");
		              else
		              {
		              printf("\nEnd of sound reached.");
		              playsel(startpos);
		              }
		             }
		if(c == 'T') {startpos=seektag(startpos);
		              printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
		              if (startpos>=prevframe(filesize))
		              printf("  End of file reached     \n");
		              else playsel(startpos);}
		if(c == 'W') {outpoint=startpos;printf("  endpoint set to %u:%05.2f  ",showmins(outpoint),showsecs(outpoint));
		              savewithtag();
		              inpoint=startpos;printf("  startpoint set to %u:%05.2f  \n",showmins(inpoint),showsecs(inpoint));}
		if(c == 'w') {outpoint=startpos;printf("  endpoint set to %u:%05.2f  \n",showmins(outpoint),showsecs(outpoint));
		              savesel();
		              inpoint=startpos;printf("  startpoint set to %u:%05.2f  \n",showmins(inpoint),showsecs(inpoint));}
		if(c == 'A') {startpos=inpoint;playsel(startpos);}
		if(c == 'B') {startpos=outpoint;playsel(startpos);}
		if(c == 's') {savesel();}
		if(c == 'S') {writeconf();}
		if(c == 't') {savewithtag();}
		if(c == 'h') help();
		if(c == '+') {silvol=silvol*1.1+1; if (silvol>10000)silvol=10000;
		              printf("  maximum volume during silence is %u  \n",silvol);}
		if(c == '-') {silvol=silvol/1.1; if (silvol<1)silvol=1;
		              printf("  maximum volume during silence is %u  \n",silvol);}
		if(c == 'm') {silencelength=silencelength*1.1; if (silencelength>10000)silencelength=10000;
		              printf("  length of silence is %u milliseconds \n",silencelength);}
		if(c == 'n') {silencelength=silencelength/1.1; if (silencelength<100)silencelength=100;
		              printf("  length of silence is %u milliseconds \n",silencelength);}
	}

	exit(0);
}

/*

MP3-Header: 4 Bytes

########
0 - 7	Alles 1 fr Sync
########
8 -10	Alles 1 fr Sync
11-12	MPEG-Version: 10=MPEG2 11=MPEG1 01=MPEG2.5
13-14	0 1 fr Layer3
	1 0 fr Layer2
	1 1 fr Layer1
   15	Protection Bit, bei 0 soll nach Header Error-Check kommen
########
16-19	Bitrate encoding
20-21	Sampling Frequency
   22	Padding Bit, if set, frame is 1 byte bigger
   23	Private Bit
########
24-25	Mono/Stereo Modus
26-27	Mode Extension fr Joint Stereo
   28	Copyright Bit
   29	Original?
30-31	Emphasis
########

*/
